@@ -149,50 +149,39 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
                    << lo << ", " << hi << ")";
     };
 
-#define S   sym.get_addr(ctx)
-#define A   rel.r_addend
-#define P   (get_addr() + rel.r_offset)
-#define G   (sym.get_got_idx(ctx) * sizeof(Word<E>))
-#define GOT ctx.got->shdr.sh_addr
+    u64 S = sym.get_addr(ctx);
+    u64 A = rel.r_addend;
+    u64 P = get_addr() + rel.r_offset;
+    u64 G = sym.get_got_idx(ctx) * sizeof(Word<E>);
+    u64 GOT = ctx.got->shdr.sh_addr;
+    u64 TOC = ctx.extra.TOC->value;
 
     switch (rel.r_type) {
     case R_PPC64_ADDR64:
-      if (sym.is_ifunc()) {
-        *dynrel++ = ElfRel<E>(P, E::R_IRELATIVE, 0, S + A);
-        if (ctx.arg.apply_dynamic_relocs)
-          *(ub64 *)loc = S + A;
-      } else {
-        apply_toc_rel(ctx, sym, rel, loc, S, A, P, dynrel);
-      }
+      apply_toc_rel(ctx, sym, rel, loc, S, A, P, dynrel);
       break;
     case R_PPC64_TOC:
-      apply_toc_rel(ctx, *ctx.TOC, rel, loc, ctx.TOC->value, A, P, dynrel);
+      apply_toc_rel(ctx, *ctx.extra.TOC, rel, loc, TOC, A, P, dynrel);
       break;
     case R_PPC64_TOC16_HA:
-      *(ub16 *)loc = ha(S + A - ctx.TOC->value);
+      *(ub16 *)loc = ha(S + A - TOC);
       break;
     case R_PPC64_TOC16_LO:
-      *(ub16 *)loc = S + A - ctx.TOC->value;
+      *(ub16 *)loc = lo(S + A - TOC);
       break;
     case R_PPC64_TOC16_DS: {
-      i64 val = S + A - ctx.TOC->value;
+      i64 val = S + A - TOC;
       check(val, -(1 << 15), 1 << 15);
       *(ub16 *)loc |= val & 0xfffc;
       break;
     }
     case R_PPC64_TOC16_LO_DS:
-      *(ub16 *)loc |= (S + A - ctx.TOC->value) & 0xfffc;
+      *(ub16 *)loc |= (S + A - TOC) & 0xfffc;
       break;
     case R_PPC64_REL24: {
       i64 val = sym.get_addr(ctx, NO_OPD) + A - P;
-
-      if (sym.has_plt(ctx) || sign_extend(val, 25) != val) {
-        RangeExtensionRef ref = extra.range_extn[i];
-        assert(ref.thunk_idx != -1);
-
-        RangeExtensionThunk<E> &thunk = *output_section->thunks[ref.thunk_idx];
-        val = thunk.get_addr(ref.sym_idx) + A - P;
-      }
+      if (sym.has_plt(ctx) || sign_extend(val, 25) != val)
+        val = get_thunk_addr(i) + A - P;
 
       check(val, -(1 << 25), 1 << 25);
       *(ub32 *)loc |= bits(val, 25, 2) << 2;
@@ -212,34 +201,34 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
       *(ub16 *)loc = ha(S + A - P);
       break;
     case R_PPC64_REL16_LO:
-      *(ub16 *)loc = S + A - P;
+      *(ub16 *)loc = lo(S + A - P);
       break;
     case R_PPC64_PLT16_HA:
-      *(ub16 *)loc = ha(G + GOT - ctx.TOC->value);
+      *(ub16 *)loc = ha(G + GOT - TOC);
       break;
     case R_PPC64_PLT16_HI:
-      *(ub16 *)loc = hi(G + GOT - ctx.TOC->value);
+      *(ub16 *)loc = hi(G + GOT - TOC);
       break;
     case R_PPC64_PLT16_LO:
-      *(ub16 *)loc = lo(G + GOT - ctx.TOC->value);
+      *(ub16 *)loc = lo(G + GOT - TOC);
       break;
     case R_PPC64_PLT16_LO_DS:
-      *(ub16 *)loc |= (G + GOT - ctx.TOC->value) & 0xfffc;
+      *(ub16 *)loc |= (G + GOT - TOC) & 0xfffc;
       break;
     case R_PPC64_GOT_TPREL16_HA:
-      *(ub16 *)loc = ha(sym.get_gottp_addr(ctx) - ctx.TOC->value);
+      *(ub16 *)loc = ha(sym.get_gottp_addr(ctx) - TOC);
       break;
     case R_PPC64_GOT_TLSGD16_HA:
-      *(ub16 *)loc = ha(sym.get_tlsgd_addr(ctx) - ctx.TOC->value);
+      *(ub16 *)loc = ha(sym.get_tlsgd_addr(ctx) - TOC);
       break;
     case R_PPC64_GOT_TLSGD16_LO:
-      *(ub16 *)loc = sym.get_tlsgd_addr(ctx) - ctx.TOC->value;
+      *(ub16 *)loc = lo(sym.get_tlsgd_addr(ctx) - TOC);
       break;
     case R_PPC64_GOT_TLSLD16_HA:
-      *(ub16 *)loc = ha(ctx.got->get_tlsld_addr(ctx) - ctx.TOC->value);
+      *(ub16 *)loc = ha(ctx.got->get_tlsld_addr(ctx) - TOC);
       break;
     case R_PPC64_GOT_TLSLD16_LO:
-      *(ub16 *)loc = ctx.got->get_tlsld_addr(ctx) - ctx.TOC->value;
+      *(ub16 *)loc = lo(ctx.got->get_tlsld_addr(ctx) - TOC);
       break;
     case R_PPC64_DTPREL16_HA:
       *(ub16 *)loc = ha(S + A - ctx.dtp_addr);
@@ -248,13 +237,13 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
       *(ub16 *)loc = ha(S + A - ctx.tp_addr);
       break;
     case R_PPC64_DTPREL16_LO:
-      *(ub16 *)loc = S + A - ctx.dtp_addr;
+      *(ub16 *)loc = lo(S + A - ctx.dtp_addr);
       break;
     case R_PPC64_TPREL16_LO:
-      *(ub16 *)loc = S + A - ctx.tp_addr;
+      *(ub16 *)loc = lo(S + A - ctx.tp_addr);
       break;
     case R_PPC64_GOT_TPREL16_LO_DS:
-      *(ub16 *)loc |= (sym.get_gottp_addr(ctx) - ctx.TOC->value) & 0xfffc;
+      *(ub16 *)loc |= (sym.get_gottp_addr(ctx) - TOC) & 0xfffc;
       break;
     case R_PPC64_PLTSEQ:
     case R_PPC64_PLTCALL:
@@ -263,14 +252,8 @@ void InputSection<E>::apply_reloc_alloc(Context<E> &ctx, u8 *base) {
     case R_PPC64_TLSLD:
       break;
     default:
-      Fatal(ctx) << *this << ": apply_reloc_alloc relocation: " << rel;
+      unreachable();
     }
-
-#undef S
-#undef A
-#undef P
-#undef G
-#undef GOT
   }
 }
 
@@ -302,8 +285,8 @@ void InputSection<E>::apply_reloc_nonalloc(Context<E> &ctx, u8 *base) {
     i64 frag_addend;
     std::tie(frag, frag_addend) = get_fragment(ctx, rel);
 
-#define S (frag ? frag->get_addr(ctx) : sym.get_addr(ctx))
-#define A (frag ? frag_addend : (i64)rel.r_addend)
+    u64 S = frag ? frag->get_addr(ctx) : sym.get_addr(ctx);
+    u64 A = frag ? frag_addend : (i64)rel.r_addend;
 
     switch (rel.r_type) {
     case R_PPC64_ADDR64:
@@ -322,11 +305,9 @@ void InputSection<E>::apply_reloc_nonalloc(Context<E> &ctx, u8 *base) {
       *(ub64 *)loc = S + A - ctx.dtp_addr;
       break;
     default:
-      Fatal(ctx) << *this << ": apply_reloc_nonalloc: " << rel;
+      Fatal(ctx) << *this << ": invalid relocation for non-allocated sections: "
+                 << rel;
     }
-
-#undef S
-#undef A
   }
 }
 
@@ -351,18 +332,17 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
     }
 
     if (sym.is_ifunc())
-      sym.flags.fetch_or(NEEDS_GOT | NEEDS_PLT | NEEDS_OPD,
+      sym.flags.fetch_or(NEEDS_GOT | NEEDS_PLT | NEEDS_PPC_OPD,
                          std::memory_order_relaxed);
 
+    // Any relocation except R_PPC64_REL24 is considered as an
+    // address-taking relocation.
     if (rel.r_type != R_PPC64_REL24 && sym.get_type() == STT_FUNC)
-      sym.flags.fetch_or(NEEDS_OPD, std::memory_order_relaxed);
+      sym.flags.fetch_or(NEEDS_PPC_OPD, std::memory_order_relaxed);
 
     switch (rel.r_type) {
     case R_PPC64_ADDR64:
-      if (sym.is_ifunc())
-        this->file.num_dynrel++;
-      else
-        scan_toc_rel(ctx, sym, rel);
+      scan_toc_rel(ctx, sym, rel);
       break;
     case R_PPC64_TOC:
       scan_toc_rel(ctx, sym, rel);
@@ -407,7 +387,7 @@ void InputSection<E>::scan_relocations(Context<E> &ctx) {
     case R_PPC64_DTPREL16_LO:
       break;
     default:
-      Fatal(ctx) << *this << ": scan_relocations: " << rel;
+      Error(ctx) << *this << ": unknown relocation: " << rel;
     }
   }
 }
@@ -422,11 +402,14 @@ void RangeExtensionThunk<E>::copy_buf(Context<E> &ctx) {
   static const ub32 pltgot_thunk[] = {
     // Store the caller's %r2
     0xf841'0028, // std   %r2, 40(%r1)
+
     // Load an address of a function descriptor
     0x3d82'0000, // addis %r12, %r2,  foo@got@toc@ha
     0xe98c'0000, // ld    %r12, foo@got@toc@lo(%r12)
+
     // Restore the callee's %r2
     0xe84c'0008, // ld    %r2,  8(%r12)
+
     // Jump to the function
     0xe98c'0000, // ld    %r12, 0(%r12)
     0x7d89'03a6, // mtctr %r12
@@ -437,11 +420,14 @@ void RangeExtensionThunk<E>::copy_buf(Context<E> &ctx) {
   static const ub32 plt_thunk[] = {
     // Store the caller's %r2
     0xf841'0028, // std   %r2, 40(%r1)
+
     // Materialize an address of a function descriptor
     0x3d82'0000, // addis %r12, %r2,  foo@gotplt@toc@ha
     0x398c'0000, // addi  %r12, %r12, foo@gotplt@toc@lo
+
     // Restore the callee's %r2
     0xe84c'0008, // ld    %r2,  8(%r12)
+
     // Jump to the function
     0xe98c'0000, // ld    %r12, 0(%r12)
     0x7d89'03a6, // mtctr %r12
@@ -470,17 +456,17 @@ void RangeExtensionThunk<E>::copy_buf(Context<E> &ctx) {
 
     if (sym.has_got(ctx)) {
       memcpy(loc, pltgot_thunk, sizeof(pltgot_thunk));
-      i64 val = sym.get_got_addr(ctx) - ctx.TOC->value;
+      i64 val = sym.get_got_addr(ctx) - ctx.extra.TOC->value;
       loc[1] |= higha(val);
       loc[2] |= lo(val);
     } else if(sym.has_plt(ctx)) {
       memcpy(loc, plt_thunk, sizeof(plt_thunk));
-      i64 val = sym.get_gotplt_addr(ctx) - ctx.TOC->value;
+      i64 val = sym.get_gotplt_addr(ctx) - ctx.extra.TOC->value;
       loc[1] |= higha(val);
       loc[2] |= lo(val);
     } else {
       memcpy(loc, local_thunk, sizeof(local_thunk));
-      i64 val = sym.get_addr(ctx, NO_OPD) - ctx.TOC->value;
+      i64 val = sym.get_addr(ctx, NO_OPD) - ctx.extra.TOC->value;
       loc[0] |= higha(val);
       loc[1] |= lo(val);
     }
@@ -511,17 +497,15 @@ get_relocation_at(Context<E> &ctx, InputSection<E> &isec, i64 offset) {
 }
 
 struct OpdSymbol {
+  bool operator<(const OpdSymbol &x) const { return r_offset < x.r_offset; }
+
   u64 r_offset = 0;
   Symbol<E> *sym = nullptr;
 };
 
 static Symbol<E> *
-get_opd_sym_at(Context<E> &ctx, std::span<OpdSymbol> syms, i64 offset) {
-  auto it = std::lower_bound(syms.begin(), syms.end(), offset,
-                             [](const OpdSymbol &ent, i64 offset) {
-    return ent.r_offset < offset;
-  });
-
+get_opd_sym_at(Context<E> &ctx, std::span<OpdSymbol> syms, u64 offset) {
+  auto it = std::lower_bound(syms.begin(), syms.end(), OpdSymbol{offset});
   if (it == syms.end())
     return nullptr;
   if (it->r_offset != offset)
@@ -567,7 +551,7 @@ get_opd_sym_at(Context<E> &ctx, std::span<OpdSymbol> syms, i64 offset) {
 // section so that they refer function symbols instead. We then mark input
 // .opd sections as dead.
 //
-// After this function, we mark symbols with the NEEDS_OPD flag if the
+// After this function, we mark symbols with the NEEDS_PPC_OPD flag if the
 // symbol needs an .opd entry. We then create an output .opd just like we
 // do for .plt or .got.
 void ppc64v1_rewrite_opd(Context<E> &ctx) {
@@ -590,7 +574,7 @@ void ppc64v1_rewrite_opd(Context<E> &ctx) {
       ElfRel<E> *rel = get_relocation_at(ctx, *opd, sym->value);
       if (!rel)
         Fatal(ctx) << *file << ": cannot find a relocation in .opd for "
-                   << *sym << " at offset 0x" << std::hex << (u32)sym->value;
+                   << *sym << " at offset 0x" << std::hex << (u64)sym->value;
 
       Symbol<E> *sym2 = file->symbols[rel->r_sym];
       if (sym2->get_type() != STT_SECTION)
@@ -603,11 +587,9 @@ void ppc64v1_rewrite_opd(Context<E> &ctx) {
     }
 
     // Sort symbols so that get_opd_sym_at() can do binary search.
-    sort(opd_syms, [](const OpdSymbol &a, const OpdSymbol &b) {
-      return a.r_offset < b.r_offset;
-    });
+    sort(opd_syms);
 
-    // Rewrite relocations directly referring .opd.
+    // Rewrite relocations so that they directly refer to .opd.
     for (std::unique_ptr<InputSection<E>> &isec : file->sections) {
       if (!isec || !isec->is_alive || isec.get() == opd)
         continue;
@@ -620,7 +602,7 @@ void ppc64v1_rewrite_opd(Context<E> &ctx) {
         Symbol<E> *real_sym = get_opd_sym_at(ctx, opd_syms, r.r_addend);
         if (!real_sym)
           Fatal(ctx) << *isec << ": cannot find a symbol in .opd for " << r
-                     << " at offset 0x" << std::hex << (u32)r.r_addend;
+                     << " at offset 0x" << std::hex << (u64)r.r_addend;
 
         r.r_sym = real_sym->sym_idx;
         r.r_addend = 0;
@@ -631,20 +613,20 @@ void ppc64v1_rewrite_opd(Context<E> &ctx) {
 
 // When a function is exported, the dynamic symbol for the function should
 // refers the function's .opd entry. This function marks such symbols with
-// NEEDS_OPD.
+// NEEDS_PPC_OPD.
 void ppc64v1_scan_symbols(Context<E> &ctx) {
   tbb::parallel_for_each(ctx.objs, [&](ObjectFile<E> *file) {
     for (Symbol<E> *sym : file->symbols)
       if (sym->file == file && sym->is_exported)
         if (u32 ty = sym->get_type(); ty == STT_FUNC || ty == STT_GNU_IFUNC)
-          sym->flags |= NEEDS_OPD;
+          sym->flags |= NEEDS_PPC_OPD;
   });
 
   // Functions referenced by the ELF header also have to have .opd entries.
   auto mark = [&](std::string_view name) {
     if (!name.empty())
       if (Symbol<E> &sym = *get_symbol(ctx, name); !sym.is_imported)
-        sym.flags |= NEEDS_OPD;
+        sym.flags |= NEEDS_PPC_OPD;
   };
 
   mark(ctx.arg.entry);
@@ -663,7 +645,7 @@ void PPC64OpdSection::copy_buf(Context<E> &ctx) {
 
   for (Symbol<E> *sym : symbols) {
     *buf++ = sym->get_addr(ctx, NO_PLT | NO_OPD);
-    *buf++ = ctx.TOC->value;
+    *buf++ = ctx.extra.TOC->value;
     *buf++ = 0;
   }
 }
