@@ -197,8 +197,8 @@ Options:
     -z notext
     -z textoff
 
-mold: supported targets: elf32-i386 elf64-x86-64 elf32-littlearm elf64-littleaarch64 elf32-littleriscv elf32-bigriscv elf64-littleriscv elf64-bigriscv elf64-powerpc elf64-powerpc elf64-powerpcle elf64-s390 elf64-sparc elf32-m68k
-mold: supported emulations: elf_i386 elf_x86_64 armelf_linux_eabi aarch64linux aarch64elf elf32lriscv elf32briscv elf64lriscv elf64briscv elf64ppc elf64lppc elf64_s390 elf64_sparc m68kelf)";
+mold: supported targets: elf32-i386 elf64-x86-64 elf32-littlearm elf64-littleaarch64 elf32-littleriscv elf32-bigriscv elf64-littleriscv elf64-bigriscv elf32-powerpc elf64-powerpc elf64-powerpc elf64-powerpcle elf64-s390 elf64-sparc elf32-m68k elf32-sh-linux elf64-alpha
+mold: supported emulations: elf_i386 elf_x86_64 armelf_linux_eabi aarch64linux aarch64elf elf32lriscv elf32briscv elf64lriscv elf64briscv elf32ppc elf32ppclinux elf64ppc elf64lppc elf64_s390 elf64_sparc m68kelf shlelf_linux elf64alpha)";
 
 static std::vector<std::string> add_dashes(std::string name) {
   // Single-letter option
@@ -519,36 +519,42 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
                    << "\n  Supported emulations:\n   elf_x86_64\n   elf_i386\n"
                    << "   aarch64linux\n   armelf_linux_eabi\n   elf64lriscv\n"
                    << "   elf64briscv\n   elf32lriscv\n   elf32briscv\n"
-                   << "   elf64ppc\n   elf64lppc\n   elf64_s390\n   elf64_sparc\n"
-                   << "   m68kelf";
+                   << "   elf32ppc\n   elf64ppc\n   elf64lppc\n   elf64_s390\n"
+                   << "   elf64_sparc\n   m68kelf\n   shlelf_linux\n   elf64alpha";
       version_shown = true;
     } else if (read_arg("m")) {
       if (arg == "elf_x86_64") {
-        ctx.arg.emulation = MachineType::X86_64;
+        ctx.arg.emulation = X86_64::target_name;
       } else if (arg == "elf_i386") {
-        ctx.arg.emulation = MachineType::I386;
+        ctx.arg.emulation = I386::target_name;
       } else if (arg == "aarch64linux") {
-        ctx.arg.emulation = MachineType::ARM64;
+        ctx.arg.emulation = ARM64::target_name;
       } else if (arg == "armelf_linux_eabi") {
-        ctx.arg.emulation = MachineType::ARM32;
+        ctx.arg.emulation = ARM32::target_name;
       } else if (arg == "elf64lriscv") {
-        ctx.arg.emulation = MachineType::RV64LE;
+        ctx.arg.emulation = RV64LE::target_name;
       } else if (arg == "elf64briscv") {
-        ctx.arg.emulation = MachineType::RV64BE;
+        ctx.arg.emulation = RV64BE::target_name;
       } else if (arg == "elf32lriscv") {
-        ctx.arg.emulation = MachineType::RV32LE;
+        ctx.arg.emulation = RV32LE::target_name;
       } else if (arg == "elf32briscv") {
-        ctx.arg.emulation = MachineType::RV32BE;
+        ctx.arg.emulation = RV32BE::target_name;
+      } else if (arg == "elf32ppc" || arg == "elf32ppclinux") {
+        ctx.arg.emulation = PPC32::target_name;
       } else if (arg == "elf64ppc") {
-        ctx.arg.emulation = MachineType::PPC64V1;
+        ctx.arg.emulation = PPC64V1::target_name;
       } else if (arg == "elf64lppc") {
-        ctx.arg.emulation = MachineType::PPC64V2;
+        ctx.arg.emulation = PPC64V2::target_name;
       } else if (arg == "elf64_s390") {
-        ctx.arg.emulation = MachineType::S390X;
+        ctx.arg.emulation = S390X::target_name;
       } else if (arg == "elf64_sparc") {
-        ctx.arg.emulation = MachineType::SPARC64;
+        ctx.arg.emulation = SPARC64::target_name;
       } else if (arg == "m68kelf") {
-        ctx.arg.emulation = MachineType::M68K;
+        ctx.arg.emulation = M68K::target_name;
+      } else if (arg == "shlelf_linux") {
+        ctx.arg.emulation = SH4::target_name;
+      } else if (arg == "elf64alpha") {
+        ctx.arg.emulation = ALPHA::target_name;
       } else {
         Fatal(ctx) << "unknown -m argument: " << arg;
       }
@@ -1071,6 +1077,7 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
     } else if (read_flag("warn-constructors")) {
     } else if (read_flag("warn-execstack")) {
     } else if (read_flag("no-warn-execstack")) {
+    } else if (read_flag("secure-plt")) {
     } else if (read_arg("rpath-link")) {
     } else if (read_z_flag("combreloc")) {
     } else if (read_z_flag("nocombreloc")) {
@@ -1175,9 +1182,9 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
       Fatal(ctx) << "-auxiliary may not be used without -shared";
   }
 
-  if (!ctx.arg.apply_dynamic_relocs && !is_rela<E>)
+  if (!ctx.arg.apply_dynamic_relocs && !E::is_rela)
     Fatal(ctx) << "--no-apply-dynamic-relocs may not be used on "
-               << E::machine_type;
+               << E::target_name;
 
   if (is_sparc<E> && ctx.arg.apply_dynamic_relocs)
     Fatal(ctx) << "--apply-dynamic-relocs may not be used on SPARC64";
