@@ -42,7 +42,7 @@ $QEMU $t/exe1 | grep -q '1 2 3 4 5 6'
 $CC -B. -o $t/exe2 $t/a.o $t/d.so $t/e.so -Wl,-no-relax
 $QEMU $t/exe2 | grep -q '1 2 3 4 5 6'
 
-if echo 'int main() {}' | $CC -o /dev/null -xc - -static >& /dev/null; then
+if test_cflags -static; then
   $CC -B. -o $t/exe3 $t/a.o $t/b.o $t/c.o -static
   $QEMU $t/exe3 | grep -q '1 2 3 4 5 6'
 fi
