@@ -1,16 +1,5 @@
 #!/bin/bash
-export LC_ALL=C
-set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
-testname=$(basename "$0" .sh)
-echo -n "Testing $testname ... "
-t=out/test/elf/$MACHINE/$testname
-mkdir -p $t
+. $(dirname $0)/common.inc
 
 cat <<EOF | $CC -fPIC -c -o $t/a.o -xc -
 void foo();
@@ -18,7 +7,7 @@ void bar() { foo(); }
 EOF
 
 $CC -B. -shared -o $t/b.so $t/a.o
-$CC -B. -shared -o $t/b.so $t/a.o -Wl,-z,nodefs
+$CC -B. -shared -o $t/b.so $t/a.o -Wl,-z,undefs
 
 ! $CC -B. -shared -o $t/b.so $t/a.o -Wl,-z,defs \
   2> $t/log || false
@@ -32,5 +21,3 @@ $CC -B. -shared -o $t/c.so $t/a.o -Wl,-z,defs \
   -Wl,--warn-unresolved-symbols 2> $t/log
 grep -q 'undefined symbol:.* foo$' $t/log
 readelf --dyn-syms $t/c.so | grep -q ' foo$'
-
-echo OK
